from django import forms

from apis_core.relations.forms import RelationForm


class LivesInForm(RelationForm):
    # we add a big notes field to this relation, both to demonstrate how custom forms
    # work in APIS, but also to test for an overflow in the form
    notes = forms.CharField(widget=forms.Textarea(attrs={"rows": "70"}), required=False)



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

from django.test import TestCase

from .helpers import construct_lookup

lookups = {
    "*foo*": ("__icontains", "foo"),
    "foo": ("__icontains", "foo"),
    "*foo": ("__iendswith", "foo"),
    "foo*": ("__istartswith", "foo"),
    '"foo"': ("__iexact", "foo"),
}


class FilterMethodstest(TestCase):
    def test_lookup(self):
        for lookup in lookups:
            self.assertEqual(construct_lookup(lookup), lookups[lookup])



# Generated by Django 5.1.1 on 2024-10-05 14:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("collections", "0002_alter_skoscollection_options"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="skoscollection",
            constraint=models.CheckConstraint(
                condition=models.Q(("name__contains", "|"), _negated=True),
                name="check_name_pipe",
                violation_error_message="The name must not contain the pipe symbol: |",
            ),
        ),
    ]



from rest_framework.routers import DefaultRouter

from apis_core.apis_metainfo.viewsets import UriToObjectViewSet

router = DefaultRouter()

router.register(r"metainfo/uritoobject", UriToObjectViewSet, basename="uritoobject")



from datetime import datetime, timedelta

from django.test import TestCase

from sample_project.models import Person, Place, Profession


class SimpleHistoryTestCase(TestCase):
    """Test of the simple_history package using the demo project"""

    def setUp(self):
        self.Person = Person
        self.Place = Place
        self.Profession = Profession

        Person.objects.create(forename="John", surname="Doe")
        Place.objects.create(
            label="Steyr", _history_date=datetime.now() - timedelta(hours=0, minutes=50)
        )

    def test_history(self):
        """Tests the simple version of attributes changed on a models instance."""
        pers = self.Person.objects.get(forename="John")
        pers.forename = "Jane"
        pers.save()
        self.assertEqual(pers.forename, "Jane")
        pers_history = pers.history.all()
        assert len(pers_history) == 2
        self.assertEqual("Jane", pers.history.most_recent().forename)
        self.assertEqual("John", pers.history.earliest().forename)

    def test_history_delete_entry(self):
        """Tests the deletion of an entry."""
        pers = self.Person.objects.get(forename="John")
        pers.delete()
        assert len(self.Person.history.all()) == 2

    def test_history_merge(self):
        """Tests the merge function of the Place model. This is still expected to fail."""
        pl1 = self.Place.objects.first()
        pl2 = self.Place.objects.create(
            label="Test", _history_date=datetime.now() - timedelta(hours=0, minutes=10)
        )
        pl1.merge_with([pl2])
        print("save()")

    def test_m2m_save(self):
        """Test if m2m profession is saved correctly."""
        pers = self.Person.objects.create(
            forename="John",
            _history_date=datetime.now() - timedelta(hours=0, minutes=10),
        )
        self.assertEqual(pers.forename, "John")
        pers_history = pers.history.all()
        assert len(pers_history) == 1
        prof = self.Profession.objects.create(name="Test")
        pers.profession.add(prof)
        pers.forename = "Jane"
        pers.save()
        assert len(pers.profession.all()) == 1
        assert len(pers.history.latest().profession.all()) == 1
        assert len(pers.history.earliest().profession.all()) == 0

    def test_history_date(self):
        """Test that history is set correctly when not manually set."""
        pers = self.Person.objects.all().first()
        pers.forename = "Jane"
        pers.save()
        assert pers.history.earliest().history_date < pers.history.latest().history_date



# Generated by Django 5.1.5 on 2025-02-27 09:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("sample_project", "0004_remove_versiongroup_self_contenttype_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="place",
            name="feature_code",
            field=models.CharField(
                blank=True,
                default="",
                help_text='<a href="https://www.geonames.org/export/codes.html">Geonames Feature Code List</a>',
                max_length=16,
            ),
        ),
        migrations.AddField(
            model_name="versionplace",
            name="feature_code",
            field=models.CharField(
                blank=True,
                default="",
                help_text='<a href="https://www.geonames.org/export/codes.html">Geonames Feature Code List</a>',
                max_length=16,
            ),
        ),
    ]



import json
import logging

import httpx
from django.template.loader import render_to_string

logger = logging.getLogger(__name__)


class ExternalAutocomplete:
    """
    This is a helper base class for implementing external
    autocomplete classes. <Modelname>ExernalAutocomplete classes
    are expected to have a `get_results(self, q)` method that
    returns a list of results usable by the autocomplete view.
    This base class implements this `get_results` method in a
    way that you can inherit from it and just define a list of
    `adapters`. Those adapters are then used one by one to
    add external autocomplete search results.
    """

    client = httpx.Client()
    adapters = []

    def get_results(self, q):
        results = []
        for adapter in self.adapters:
            results.extend(adapter.get_results(q, self.client))
        return results


class ExternalAutocompleteAdapter:
    """
    Base class for ExternalAutocompleteAdapters. It provides
    the methods used for templating the autocomplete results.
    You can pass a `template` name to initialization, which
    is then used to style the results.
    """

    template = None

    def __init__(self, *args, **kwargs):
        self.template = kwargs.get("template", None)

    def default_template(self, result):
        return f'{result["label"]} <a href="{result["id"]}">{result["id"]}</a>'

    def get_result_label(self, result):
        if self.template:
            return render_to_string(self.template, {"result": result})
        return self.default_template(result)


class TypeSenseAutocompleteAdapter(ExternalAutocompleteAdapter):
    """
    This autocomplete adapters queries typesense collections on a
    typesense server. The `collections` variable can either be a
    string or a list - if its a string, that collection is queried
    directly, if its a list, the adapter uses typesense `multi_search`
    endpoint.
    """

    collections = None
    token = None
    server = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.collections = kwargs.get("collections", None)
        self.token = kwargs.get("token", None)
        self.server = kwargs.get("server", None)

    def default_template(self, result):
        return super().default_template(result["document"])

    def extract(self, res):
        if res.get("document"):
            return {
                "id": res["document"]["id"],
                "text": self.get_result_label(res),
                "selected_text": self.get_result_label(res),
            }
        logger.error(
            "Could not parse result from typesense collection %s: %s",
            self.collections,
            res,
        )
        return False

    def get_results(self, q, client=httpx.Client()):
        headers = {"X-TYPESENSE-API-KEY": self.token}
        res = None
        if self.token and self.server:
            match self.collections:
                # if there is only on collection configured, we hit that collection directly
                case str() as collection:
                    url = f"{self.server}/collections/{collection}/documents/search?q={q}&query_by=description&query_by=label"
                    res = client.get(url, headers=headers)
                # if there are multiple collections configured, we use the `multi_search` endpoint
                case list() as collectionlist:
                    url = f"{self.server}/multi_search?q={q}&query_by=description&query_by=label"
                    data = {"searches": []}
                    for collection in collectionlist:
                        data["searches"].append({"collection": collection})
                    res = client.post(url, data=json.dumps(data), headers=headers)
                case unknown:
                    logger.error("Don't know what to do with collection %s", unknown)

            if res:
                data = res.json()
                hits = data.get("hits", [])
                for result in data.get("results", []):
                    hits.extend(result["hits"])
                return list(filter(bool, map(self.extract, hits)))
        return []


class LobidAutocompleteAdapter(ExternalAutocompleteAdapter):
    """
    This autocomplete adapters queries the lobid autocomplete apis.
    See https://lobid.org/gnd/api for details
    You can pass a `lobid_params` dict which will then be use as GET
    request parameters.
    """

    params = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params = kwargs.get("params", {})

    def extract(self, res):
        return {
            "id": res["id"],
            "text": self.get_result_label(res),
            "selected_text": self.get_result_label(res),
        }

    def get_results(self, q, client=httpx.Client()):
        endpoint = "https://lobid.org/gnd/search?"
        self.params["q"] = q
        res = client.get(endpoint, params=self.params)
        if res:
            return list(filter(bool, map(self.extract, res.json())))
        return []



import json
import logging
import urllib
from functools import cache

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db.utils import IntegrityError

from apis_core.apis_metainfo.models import Uri
from apis_core.apis_metainfo.utils import create_object_from_uri
from apis_core.utils.rdf import get_something_from_uri

logger = logging.getLogger(__name__)


class GenericModelImporter:
    """
    A generic importer class which provides methods for
    importing data from a URI and creating a model instance from it.

    By default, it fetches a resource and first tries to parse it using
    our RDF parser. If that fails, it tries to parse it using JSON and
    then extracts the fields whose keys match the model field names.
    Projects can inherit from this class and override the default
    methods or simply write their own from scratch.
    """

    model = None
    import_uri = None

    def __init__(self, uri, model):
        self.model = model
        self.import_uri = self.clean_uri(uri)

    @property
    def get_uri(self):
        return self.import_uri

    def clean_uri(self, uri):
        return get_normalized_uri(uri)

    @cache
    def request(self, uri):
        # we first try to use the RDF parser
        try:
            data = get_something_from_uri(
                uri,
                [self.model],
            )
            return data
        except Exception as e:
            logger.debug(e)
        # if everything else fails, try parsing JSON
        # if even that does not help, return an empty dict
        try:
            return json.loads(urllib.request.urlopen(uri).read())
        except Exception as e:
            logger.debug(e)
        return {}

    def mangle_data(self, data):
        return data

    def get_data(self, drop_unknown_fields=True):
        """
        Fetch the data using the `request` method and
        mangle it using the `mangle_data` method.

        If the `drop_unknown_fields` argument is True,
        remove all fields from the data dict that do not
        have an equivalent field in the model.
        """
        data = self.request(self.import_uri)
        data = self.mangle_data(data)
        if drop_unknown_fields:
            # we are dropping all fields that are not part of the model
            modelfields = [field.name for field in self.model._meta.fields]
            data = {key: data[key] for key in data if key in modelfields}
        if not data:
            raise ImproperlyConfigured(
                f"Could not import {self.import_uri}. Data fetched was: {data}"
            )
        return data

    def import_into_instance(self, instance, fields="__all__"):
        data = self.get_data()
        if fields == "__all__":
            fields = data.keys()
        for field in fields:
            if hasattr(instance, field) and field in data.keys():
                setattr(instance, field, data[field][0])
        instance.save()

    def create_instance(self):
        logger.debug("Create instance from URI %s", self.import_uri)
        data = self.get_data(drop_unknown_fields=False)
        instance = None
        same_as = data.get("same_as", [])
        same_as = [get_normalized_uri(uri) for uri in same_as]
        if sa := Uri.objects.filter(uri__in=same_as):
            root_set = set([s.content_object for s in sa])
            if len(root_set) > 1:
                raise IntegrityError(
                    f"Multiple objects found for sameAs URIs {data['same_as']}. "
                    f"This indicates a data integrity problem as these URIs should be unique."
                )
            instance = sa.first().content_object
            logger.debug("Found existing instance %s", instance)
        if not instance:
            attributes = {}
            for field in self.model._meta.fields:
                if data.get(field.name, False):
                    attributes[field.name] = data[field.name][0]
            instance = self.model.objects.create(**attributes)
            logger.debug("Created instance %s from attributes %s", instance, attributes)
        content_type = ContentType.objects.get_for_model(instance)
        for uri in same_as:
            Uri.objects.get_or_create(
                uri=uri, content_type=content_type, object_id=instance.id
            )
        for relation, details in data.get("relations", {}).items():
            rel_app_label, rel_model = relation.split(".")
            relation_model = ContentType.objects.get_by_natural_key(
                app_label=rel_app_label, model=rel_model
            ).model_class()

            reld = details.get("obj", None) or details.get("subj", None)
            reld_app_label, reld_model = reld.split(".")
            related_content_type = ContentType.objects.get_by_natural_key(
                app_label=reld_app_label, model=reld_model
            )
            related_model = related_content_type.model_class()

            for related_uri in details["curies"]:
                try:
                    related_instance = create_object_from_uri(
                        uri=related_uri, model=related_model
                    )
                    if details.get("obj"):
                        subj_object_id = instance.pk
                        subj_content_type = content_type
                        obj_object_id = related_instance.pk
                        obj_content_type = related_content_type
                    else:
                        obj_object_id = instance.pk
                        obj_content_type = content_type
                        subj_object_id = related_instance.pk
                        subj_content_type = related_content_type
                    rel, _ = relation_model.objects.get_or_create(
                        subj_object_id=subj_object_id,
                        subj_content_type=subj_content_type,
                        obj_object_id=obj_object_id,
                        obj_content_type=obj_content_type,
                    )
                    logger.debug(
                        "Created relation %s between %s and %s",
                        relation_model.name(),
                        rel.subj,
                        rel.obj,
                    )
                except Exception as e:
                    logger.error(
                        "Could not create relation to %s due to %s", related_uri, e
                    )
        return instance



from django.contrib.contenttypes.models import ContentType

from apis_core.collections.models import SkosCollectionContentObject
from apis_core.generic.filtersets import GenericFilterSet


class SkosCollectionContentObjectFilterSet(GenericFilterSet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.filters["collections"]
        content_types = SkosCollectionContentObject.objects.all().values("content_type")
        self.filters["content_type"].queryset = ContentType.objects.filter(
            pk__in=content_types
        )



from apis_core.generic.forms import GenericModelForm


class UriForm(GenericModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper.form_class = "form-horizontal"
        self.helper.label_class = "col-md-3"
        self.helper.field_class = "col-md-9"



# Generated by Django 4.1.10 on 2023-09-21 07:05

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_entities", "0003_remove_tempentityclass_source"),
        ("apis_metainfo", "0006_delete_text"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Source",
        ),
    ]



import logging

import django_filters
from django.apps import apps
from django_filters.filterset import FilterSet

from apis_core.collections.filters import CollectionsIncludeExcludeFilter

from .forms import GenericFilterSetForm

logger = logging.getLogger(__name__)


class GenericFilterSet(FilterSet):
    """
    Our GenericFilterSet sets the default `form` to be our
    GenericFilterSetForm, which is set up to ignore the `columns` field
    of the form.
    """

    class Meta:
        form = GenericFilterSetForm
        # we set the UnknownFieldBehavior to WARN, so the form does not
        # break if there are JSONFields
        unknown_field_behavior = django_filters.UnknownFieldBehavior.WARN

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        model = self._meta.model
        # remove all the filters that are based on auto_created model fields
        for field in model._meta.get_fields():
            if getattr(field, "auto_created", False) and field.name in self.filters:
                del self.filters[field.name]
        try:
            skoscollection = apps.get_model("collections.SkosCollection")
            if skoscollection.objects.exists():
                self.filters["collections"] = CollectionsIncludeExcludeFilter(
                    queryset=skoscollection.objects.all(),
                )
        except LookupError as e:
            logger.debug("Not adding collections filter to form: %s", e)



# Generated by Django 5.1.7 on 2025-04-29 12:47

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("sample_project", "0007_alter_person_forename_alter_versionperson_forename"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="versiongroup",
            name="version_tag",
        ),
        migrations.RemoveField(
            model_name="versionperson",
            name="version_tag",
        ),
        migrations.RemoveField(
            model_name="versionplace",
            name="version_tag",
        ),
    ]



from django.db.models import Q
from django.template import loader
from rest_framework.filters import SearchFilter

from apis_core.apis_entities.utils import get_entity_classes
from apis_core.generic.helpers import generate_search_filter


class CustomSearchFilter(SearchFilter):
    def filter_queryset(self, request, queryset, view):
        search_terms = self.get_search_terms(request)
        if not search_terms:
            return queryset
        entities = get_entity_classes()
        q = Q()
        for entity in entities:
            name = entity._meta.model_name
            q |= Q(**{f"{name}__isnull": False}) & generate_search_filter(
                entity, search_terms, prefix=f"{name}__"
            )
        return queryset.filter(q)

    def to_html(self, request, queryset, view):
        context = {
            "param": self.search_param,
            "term": request.query_params.get(self.search_param, ""),
        }
        template = loader.get_template(self.template)
        return template.render(context)



from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from apis_core.core.fields import ApisListSelect2
from apis_core.generic.forms import GenericFilterSetForm, GenericModelForm
from apis_core.generic.forms.fields import ModelImportChoiceField


class RelationForm(GenericModelForm):
    """
    This form overrides generic form for relation editing.
    Relations have generic relations to subj and obj, but we
    hide those ForeignKey form fields and instead show
    autocomplete choices fields.
    In addition, one can pass a hx_post_route argument to the
    form to make the form set the `hx-post` attribute to the
    given value.
    We also pass a `reverse` boolean, wich gets passed on
    to the htmx POST endpoint using url parameters (the endpoint
    can then select the success_url based on the `reverse` state).
    """

    class Media:
        js = ["js/relation_dialog.js"]

    class Meta:
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        """
        Initialize the form and replace `subj_object_id` and
        `obj_object_id` fields with a autocomplete widget.
        """

        self.params = kwargs.pop("params", {})
        subj_model = getattr(self.Meta.model, "subj_model", None)
        obj_model = getattr(self.Meta.model, "obj_model", None)

        if subj_model:
            subj_ct = ContentType.objects.get_for_model(subj_model)
            kwargs["initial"]["subj_content_type"] = subj_ct
        if obj_model:
            obj_ct = ContentType.objects.get_for_model(obj_model)
            kwargs["initial"]["obj_content_type"] = obj_ct

        super().__init__(*args, **kwargs)

        if subj_model:
            self.fields["subj_content_type"].required = False
            self.fields["subj_content_type"].widget = forms.HiddenInput()
            # if we already now the id of the subject, hide the subject input
            if kwargs["initial"].get("subj_object_id"):
                self.fields["subj_object_id"].widget = forms.HiddenInput()
            # otherwise create an autocomplete to select the subject
            else:
                self.fields["subj_object_id"] = ModelImportChoiceField(
                    queryset=subj_model.objects.all(), label=_("Subject")
                )
                self.fields["subj_object_id"].widget = ApisListSelect2(
                    attrs={"data-html": True},
                    url=reverse("apis_core:generic:autocomplete", args=[subj_ct]),
                )
                self.fields["subj_object_id"].widget.choices = self.fields[
                    "subj_object_id"
                ].choices

        if obj_model:
            self.fields["obj_content_type"].required = False
            self.fields["obj_content_type"].widget = forms.HiddenInput()
            # if we already now the id of the object, hide the object input
            if kwargs["initial"].get("obj_object_id"):
                self.fields["obj_object_id"].widget = forms.HiddenInput()
            # otherwise create an autocomplete to select the object
            else:
                self.fields["obj_object_id"] = ModelImportChoiceField(
                    queryset=obj_model.objects.all(), label=_("Object")
                )
                self.fields["obj_object_id"].widget = ApisListSelect2(
                    attrs={"data-html": True},
                    url=reverse("apis_core:generic:autocomplete", args=[obj_ct]),
                )
                self.fields["obj_object_id"].widget.choices = self.fields[
                    "obj_object_id"
                ].choices

        self.order_fields(self.field_order)
        self.helper = FormHelper(self)
        model_ct = ContentType.objects.get_for_model(self.Meta.model)
        self.helper.form_id = f"relation_{model_ct.model}_form"
        self.helper.add_input(Submit("submit", "Submit"))

        if hx_post_route := self.params.get("hx_post_route", False):
            self.helper.attrs = {
                "hx-post": hx_post_route + "?" + urlencode(self.params, doseq=True),
                "hx-swap": "outerHTML",
                "hx-target": f"#{self.helper.form_id}",
            }

    def clean(self) -> dict:
        """
        subj_object_id and obj_object_id can either come
        from initial values or from the autocomplete.
        In the former case, the values are simply the primary keys,
        which is what we want.
        In the latter case, the values are object instances and we
        want to set them to the primary keys instead.
        """
        cleaned_data = super().clean()
        subj_object_id = cleaned_data.get("subj_object_id")
        if subj_object_id and not isinstance(subj_object_id, int):
            cleaned_data["subj_object_id"] = subj_object_id.id
        obj_object_id = cleaned_data.get("obj_object_id")
        if obj_object_id and not isinstance(obj_object_id, int):
            cleaned_data["obj_object_id"] = obj_object_id.id
        return cleaned_data

    @property
    def relation_name(self) -> str:
        """A helper method to access the correct name of the relation"""
        if self.params["reverse"]:
            return self._meta.model.reverse_name
        return self._meta.model.name


class RelationFilterSetForm(GenericFilterSetForm):
    """
    FilterSet form for relations based on GenericFilterSetForm.
    Excludes `relation_ptr` from the columns selector
    """

    columns_exclude = ["relation_ptr"]



# Generated by Django 4.1.10 on 2023-09-13 12:17

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_entities", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="tempentityclass",
            name="text",
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.http import urlencode
from django.views.generic.base import TemplateView
from django.views.generic.detail import BaseDetailView

from apis_core.generic.helpers import split_and_strip_parameter, string_to_bool
from apis_core.generic.templatetags.generic import content_types_by_natural_keys
from apis_core.generic.views import Create, GenericModelMixin
from apis_core.relations.templatetags.relations import (
    possible_relation_types_from,
    relations_from,
)


class CreateRelation(Create):
    def get_form_kwargs(self, *args, **kwargs) -> dict:
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["initial"]["obj_content_type"] = self.request.GET.get("obj_content_type")
        kwargs["initial"]["obj_object_id"] = self.request.GET.get("obj_object_id")
        kwargs["initial"]["subj_content_type"] = self.request.GET.get(
            "subj_content_type"
        )
        kwargs["initial"]["subj_object_id"] = self.request.GET.get("subj_object_id")
        return kwargs


class RelationParamsMixin:
    def setup(self, *args, **kwargs) -> None:
        super().setup(*args, **kwargs)
        self.params = self.request.GET.dict()
        self.params["reverse"] = string_to_bool(self.params.get("reverse", "false"))
        self.params["relation_types"] = split_and_strip_parameter(
            self.request.GET.getlist("relation_types")
        )


class CreateRelationForm(RelationParamsMixin, CreateRelation):
    template_name = "relations/create_relation_form.html"

    def get(self, *args, **kwargs):
        resp = super().get(*args, **kwargs)
        content_type = ContentType.objects.get_for_model(self.model)
        resp["HX-Trigger-After-Settle"] = (
            '{"reinit_select2": "relation_' + content_type.model + '_form"}'
        )
        return resp

    def get_form_kwargs(self, *args, **kwargs) -> dict:
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["params"] = self.params
        content_type = ContentType.objects.get_for_model(self.model)
        kwargs["params"]["hx_post_route"] = reverse(
            "apis_core:relations:create_relation_form", args=[content_type]
        )
        return kwargs

    def get_success_url(self) -> str:
        obj = self.object.obj if self.params["reverse"] else self.object.subj
        args = [obj.content_type, obj.id]
        params = {
            k: self.params[k] for k in ["relation_types", "replace_id", "table_suffix"]
        }
        params = "?" + urlencode(params, doseq=True)
        return reverse("apis_core:relations:list_relations", args=args) + params


class ListRelations(
    RelationParamsMixin, GenericModelMixin, BaseDetailView, TemplateView
):
    template_name = "relations/partials/list_relations.html"

    def get_context_data(self, *args, **kwargs) -> dict:
        ctx = super().get_context_data(*args, **kwargs)
        ctx["relations"] = []

        relation_types = content_types_by_natural_keys(
            tuple(self.params["relation_types"])
        )
        if not relation_types:
            relation_types = possible_relation_types_from(self.object)
        for relation_type in relation_types:
            ctx["relations"].extend(relations_from(self.object, relation_type))

        ctx.update(self.params)

        return ctx

    def get(self, *args, **kwargs):
        resp = super().get(*args, **kwargs)
        resp["HX-Trigger"] = "dismissModal"
        return resp



from django.contrib.contenttypes.models import ContentType
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import include, path, register_converter
from rest_framework import routers

from apis_core.generic import api_views, views

from .abc import GenericModel

app_name = "generic"


class ContenttypeConverter:
    """
    A converter that converts from a string representation of a
    model (`app_label.model`) to the actual Django model class.
    """

    regex = r"\w+\.\w+"

    def to_python(self, value):
        app_label, model = value.split(".")
        contenttype = get_object_or_404(ContentType, app_label=app_label, model=model)
        if issubclass(contenttype.model_class(), GenericModel):
            return contenttype
        raise Http404

    def to_url(self, value):
        if isinstance(value, ContentType):
            return f"{value.app_label}.{value.model}"
        if isinstance(value, str):
            return value


register_converter(ContenttypeConverter, "contenttype")

router = routers.DefaultRouter()
router.register(r"", api_views.ModelViewSet, basename="genericmodelapi")

urlpatterns = [
    path("overview/", views.Overview.as_view(), name="overview"),
    path(
        "<contenttype:contenttype>/",
        include(
            [
                path("", views.List.as_view(), name="list"),
                path("<int:pk>", views.Detail.as_view(), name="detail"),
                path("create", views.Create.as_view(), name="create"),
                path("delete/<int:pk>", views.Delete.as_view(), name="delete"),
                path("update/<int:pk>", views.Update.as_view(), name="update"),
                path("duplicate/<int:pk>", views.Duplicate.as_view(), name="duplicate"),
                path(
                    "selectmergeorenrich/<int:pk>",
                    views.SelectMergeOrEnrich.as_view(),
                    name="selectmergeorenrich",
                ),
                path(
                    "merge/<int:pk>/<int:otherpk>",
                    views.MergeWith.as_view(),
                    name="merge",
                ),
                path("enrich/<int:pk>", views.Enrich.as_view(), name="enrich"),
                path("autocomplete", views.Autocomplete.as_view(), name="autocomplete"),
                path("import", views.Import.as_view(), name="import"),
                path(
                    "autocomplete/externalonly",
                    views.Autocomplete.as_view(),
                    {"external_only": True},
                    name="autocompleteexternalonly",
                ),
            ]
        ),
    ),
    path("api/<contenttype:contenttype>/", include(router.urls)),
]



import re
from urllib.parse import urlsplit

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri, get_rules
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from model_utils.managers import InheritanceManager

from apis_core.generic.abc import GenericModel
from apis_core.utils import rdf
from apis_core.utils import settings as apis_settings


class RootObject(GenericModel, models.Model):
    """
    The very root thing that can exist in a given ontology. Several classes inherit from it.
    By having one overarching super class we gain the advantage of unique identifiers.
    """

    objects = models.Manager()
    objects_inheritance = InheritanceManager()


# Uri model
# We use a custom UriManager, so we can override the queryset `get`
# method. This way we can normalize the uri field.


class UriQuerySet(models.query.QuerySet):
    def get(self, *args, **kwargs):
        if "uri" in kwargs:
            kwargs["uri"] = get_normalized_uri(kwargs["uri"])
        return super().get(*args, **kwargs)


class UriManager(models.Manager):
    def get_queryset(self):
        return UriQuerySet(self.model)


class Uri(GenericModel, models.Model):
    """
    The URI model provides a way of storing globally scoped identifiers for
    objects. For the objects in our application, we store both internal URIs
    (URIs that we created) as well as external URIs (when we want to link the
    object to an external resource, asserting that this object is the "same as"
    the external object).
    """

    uri = models.URLField(blank=True, null=True, unique=True, max_length=255)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.PositiveIntegerField(null=True)
    content_object = GenericForeignKey()

    objects = UriManager()

    def __str__(self):
        return str(self.uri)

    def save(self, *args, **kwargs):
        self.clean()
        return super().save(*args, **kwargs)

    def clean(self):
        self.uri = get_normalized_uri(self.uri)
        if self.uri and not hasattr(self, "content_object"):
            try:
                result = rdf.get_something_from_uri(self.uri)
                if model := result.getattr("model", False):
                    obj = model(**result.get("attributes", {}))
                    obj.save()
                    self.content_type = ContentType.objects.get_for_model(obj)
                    self.object_id = obj.id
                else:
                    raise ImproperlyConfigured(
                        f"{self.uri}: did not find matching rdf defintion"
                    )
            except Exception as e:
                raise ValidationError(f"{e}: {self.uri}")

    def internal(self) -> bool:
        my_netloc = urlsplit(self.uri).netloc
        return any(
            [my_netloc == urlsplit(uri).netloc for uri in apis_settings.internal_uris()]
        )

    def short_label(self) -> str:
        if self.internal():
            return "local"
        for x in get_rules():
            if re.match(x["match"], self.uri):
                return x["name"]
        return urlsplit(self.uri).netloc



from django.core.exceptions import ValidationError
from django.forms import ModelChoiceField, MultiValueField, MultiWidget
from django.utils.translation import gettext as _

from apis_core.apis_metainfo.utils import create_object_from_uri


class ModelImportChoiceField(ModelChoiceField):
    def to_python(self, value):
        result = None
        if value.startswith(("http://", "https://")):
            try:
                result = create_object_from_uri(
                    value, self.queryset.model, raise_on_fail=True
                )
            except Exception as e:
                raise ValidationError(
                    _("Could not import %(value)s: %(exception)s"),
                    params={"value": value, "exception": e},
                )
        return result or super().to_python(value)


class RowColumnMultiWidget(MultiWidget):
    """
    A custom MultiWidget that is meant to be used with the
    RowColumnMultiValueField. The widget takes a list of widgets
    as a parameter and displays those widgets in columns in one row.
    The `labels` parameter is used to add a separate label to the
    individual widgets.
    """

    template_name = "widgets/row_column_multiwidget.html"
    use_fieldset = False

    def __init__(self, widgets, labels=[], attrs=None):
        self.labels = labels
        super().__init__(widgets, attrs)

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)
        for widget in ctx["widget"]["subwidgets"]:
            if self.labels:
                widget["label"] = self.labels.pop(0)
        return ctx

    def decompress(self, value):
        if value:
            return value
        return []


class RowColumnMultiValueField(MultiValueField):
    """
    This is a custom MultiValueField that simply shows multiple form
    fields in a row. The form fields are passed to the constructor and
    the corresponding RowColumnMultiWidget simply iterates through all
    the fields and shows them in rows.
    Additionaly it is possible to pass a list of `labels` that are then
    also passed on to the widget, which uses those to add a separate
    label to the individual widgets.
    """

    def __init__(self, fields, labels=[], *args, **kwargs):
        kwargs["widget"] = RowColumnMultiWidget(
            widgets=[f.widget for f in fields], labels=labels
        )
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return data_list



from django import template
from django.contrib.contenttypes.models import ContentType

from apis_core.collections.models import SkosCollection, SkosCollectionContentObject

register = template.Library()


### templatetags for the APIS collections model ###


@register.inclusion_tag("collections/collection_toggle.html", takes_context=True)
def collection_toggle(context, obj, collection):
    """
    Provide a button to add or remove a connection between a
    collection and an object.
    """
    content_type = ContentType.objects.get_for_model(obj)
    context["content_type"] = content_type
    context["object"] = obj
    context["collection"] = collection
    context["exists"] = SkosCollectionContentObject.objects.filter(
        object_id=obj.id, content_type=content_type, collection=collection
    ).exists()
    return context


@register.inclusion_tag("collections/collection_toggle.html", takes_context=True)
def collection_toggle_by_id(context, obj, collectionid):
    """
    Wrapper templatetag to allow using `collection_toggle`
    with just the `id` of the collection.
    """
    collection = SkosCollection.objects.get(pk=collectionid)
    return collection_toggle(context, obj, collection)


@register.inclusion_tag(
    "collections/collection_object_collection.html", takes_context=True
)
def collection_object_collection(context, obj, skoscollectioncollectionobjects):
    """
    Provide a button to change the connection between an object and
    a collection to point to the collection the collection is in.
    """
    if len(skoscollectioncollectionobjects) > 1:
        context["error"] = (
            "Multiple collections found to toggle, please check `collection_object_collection`"
        )
    if len(skoscollectioncollectionobjects) == 1:
        collectionobject = skoscollectioncollectionobjects.first()
        context["parent"] = collectionobject.collection.parent_collection
        context["collectionobject"] = collectionobject
        context["content_type"] = ContentType.objects.get_for_model(obj)
        context["object"] = obj
    return context


@register.inclusion_tag(
    "collections/collection_object_collection.html", takes_context=True
)
def collection_object_collection_by_id(context, obj, *collection_ids):
    """
    Wrapper templatetag to allow using `collection_object_parent` with
    just the `ids` of collections.
    """
    content_type = ContentType.objects.get_for_model(obj)
    sccos = SkosCollectionContentObject.objects.filter(
        collection__in=collection_ids, content_type=content_type, object_id=obj.id
    )
    return collection_object_collection(context, obj, sccos)


@register.simple_tag(takes_context=True)
def collection_content_objects(context, obj, collectionids=None):
    content_type = ContentType.objects.get_for_model(obj)
    sccos = SkosCollectionContentObject.objects.filter(
        content_type=content_type, object_id=obj.id
    )
    if collectionids is not None:
        ids = collectionids.split(",")
        sccos = sccos.filter(collection__id__in=ids)
    return sccos


@register.inclusion_tag(
    "collections/collection_session_toggle.html", takes_context=True
)
def collection_session_toggle_by_id(context, collection_id):
    """
    Provide a checkbox to toggle if a session collection is active.
    The checkbox calls the CollectionSessionToggle view.
    """
    session_collections = context.request.session.get("session_collections", [])
    context["collection"] = SkosCollection.objects.get(pk=collection_id)
    context["enabled"] = collection_id in session_collections
    return context



import logging

from django.apps import AppConfig
from django.conf import settings

logger = logging.getLogger(__name__)


class MetainfoConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "apis_core.apis_metainfo"

    def ready(self):
        from . import signals  # noqa: F401

        if getattr(settings, "APIS_BASE_URI", None) is None:
            logger.warning(
                "You should set the APIS_BASE_URI setting - we are using https://example.org as a fallback!"
            )
